//-----------------------------------------------------------
//    sym_banded_matrix.h
//-----------------------------------------------------------
#ifndef SYKES_NUMERIC_SYM_BANDED_MATRIX_H
#define SYKES_NUMERIC_SYM_BANDED_MATRIX_H

#include <stdexcept>
#include <memory>
#include <iterator>
#include <type_traits>
#include <cstddef>
#include <cassert>
#include <new>
#include "matrix_and_vector_base.h"
#include "interval_iterator.h"
#include "reverse_iterator.h"
#include "vector.h"

namespace sykes{ namespace numeric{
    //-----------------------------------------------------------
    //    class sym_banded_matrix
    //-----------------------------------------------------------
    template<typename Tp, typename Allocator = std::allocator<Tp> >
    class sym_banded_matrix : public matrix_base
    {
    public:
        typedef sym_banded_matrix self_type;
        typedef Tp value_type;
        typedef Allocator allocator_type;
        typedef typename Allocator::reference reference;
        typedef typename Allocator::const_reference const_reference;
        typedef typename Allocator::size_type size_type;
        typedef typename Allocator::pointer pointer;
        typedef typename Allocator::const_pointer const_pointer;
    
    private:
        Allocator m_allocator;
        size_type m_size;
        size_type m_half_band;
        pointer m_start;
        pointer m_finish;
        static value_type const m_zero;
        
    public:
        
        //---- constructor
        sym_banded_matrix()
            : m_allocator(), m_size(0), m_half_band(0), m_start(0), m_finish(0)
        {}
        
        sym_banded_matrix(size_type _size, size_type _band, value_type const& _val = value_type())
            : m_allocator(), m_size(_size), m_half_band(_band), m_start(0), m_finish(0)
        {
            size_type storage_size = _size * _band;
            m_start = m_allocator.allocate(storage_size);
            m_finish = m_start + storage_size;
            std::uninitialized_fill(m_start, m_finish, _val);
        }
        
        //---- copy constructor
        sym_banded_matrix(self_type const& _other)
            : m_allocator(), m_size(_other.m_size), m_half_band(_other.m_half_band),
                m_start(), m_finish()
        {
            m_start = m_allocator.allocate(_other.storage_size());
            m_finish =
                std::uninitialized_copy(_other.m_start, _other.m_finish, m_start);
        }
        
        //---- destructor
        ~sym_banded_matrix()
        {
            size_type size1 = storage_size();
            for(size_type i1 = 0; i1 < size1; ++i1)
                m_allocator.destroy(m_start + i1);
            m_allocator.deallocate(m_start, size1);
        }
        
        //---- resize
        void resize(size_type _size, size_type _band, bool _preserve = true, value_type const& _val = value_type())
        {
            size_type new_storage_size = _size * _band;
            size_type old_storage_size = storage_size();
            
            if(_size == m_size && _band == m_half_band)
                return;
            else if(new_storage_size > max_storage_size())
                throw std::length_error("sykes::sym_banded_matrix::resize()");
            else
            {
                pointer tmp = m_allocator.allocate(new_storage_size);
                std::uninitialized_fill(tmp, tmp + new_storage_size, _val);
                
                if(_preserve){
                    size_type cpband = ((_band < m_half_band)?  _band : m_half_band);
                    size_type cpsize = ((_size < m_size)?  _size : m_size);
                    for(size_type i1 = 0; i1 < cpsize; ++i1){
                        for(size_type i2 = 0; i2 < cpband; ++i2){
                            *(tmp + i1 * _band + i2) = *(m_start + i1 * m_half_band + i2);
                        }
                    }
                }
                
                for(size_type i1 = 0; i1 < old_storage_size; ++i1)
                    m_allocator.destroy(m_start + i1);
                m_allocator.deallocate(m_start, old_storage_size);
                m_size = _size;
                m_half_band = _band;
                
                m_start = tmp;
                m_finish = tmp + new_storage_size;
            }
        }
        
        //---- operator=
        self_type& operator=(self_type const& _other)
        {
            size_type new_storage_size = _other.storage_size();
            size_type old_storage_size = storage_size();
            if(new_storage_size > max_storage_size())
                throw std::length_error("sykes::sym_banded_matrix::resize()");
            else if(m_size != _other.m_size || m_half_band != _other.m_half_band){
                pointer tmp = m_allocator.allocate(new_storage_size), tmp_finish(0);
                tmp_finish = std::uninitialized_copy(_other.m_start, _other.m_finish, tmp);
                
                for(size_type i1 = 0; i1 < old_storage_size; ++i1)
                    m_allocator.destroy(m_start + i1);
                m_allocator.deallocate(m_start, old_storage_size);
                m_size = _other.m_size;
                m_half_band = _other.m_half_band;
                m_start = tmp;
                m_finish = tmp + new_storage_size;
            }else
                std::copy(_other.m_start, _other.m_finish, m_start);
            return (*this);
        }
        
        //---- element access
        reference operator()(size_type _row, size_type _col)
        {
            size_type x = _row, y = _col;
            if(x < y){
                x = _col;
                y = _row;
            }
            assert(x < m_size && (x - y < m_half_band));
            return *(m_start + y * m_half_band + x - y);
        }
        
        const_reference operator()(size_type _row, size_type _col) const
        {
            size_type x = _row, y = _col;
            if(x < y){
                x = _col;
                y = _row;
            }
            assert(x < m_size && y < m_size);
            if(x - y < m_half_band)
                return *(m_start + y * m_half_band + x - y);
            else return m_zero;
        }
        
        value_type& at(size_type _row, size_type _col)
        {
            size_type x = _row, y = _col;
            if(x < y){
                x = _col;
                y = _row;
            }
            assert(x < m_size && (y - x < m_half_band));
            if(x >= m_size || (x - y >= m_half_band)){
                throw std::range_error("sykes::numeric::sym_banded_matrix::at()");
            }
            return *(m_start + y * m_half_band + x - y);
        }
        
        value_type const& at(size_type _row, size_type _col) const
        {
            size_type x = _row, y = _col;
            if(x < y){
                x = _col;
                y = _row;
            }
            assert(x < m_size && (y - x < m_half_band));
            if(x >= m_size || (x - y >= m_half_band)){
                throw std::range_error("sykes::numeric::sym_banded_matrix::at()");
            }
            return *(m_start + y * m_half_band + x - y);
        }
        
        //---- size
        size_type size() const { return m_size; }
        size_type size1() const { return m_size; }
        size_type size2() const { return m_size; }
        size_type get_half_band_width() const { return m_half_band; }
        size_type storage_size() const { return (m_finish - m_start); }
        size_type max_storage_size() const { return m_allocator.max_size(); }
        bool empty() const { return (m_start == m_finish);}
        
        //---- get_allocator
        allocator_type get_allocator() const { return m_allocator; }
        
        //---- data()
        Tp const* data() const { return m_start; }
        Tp* data() { return m_start; }
        
        //---- comparison
        bool operator==(self_type const& _other) const
        {
            bool tmp = ((m_size == _other.m_size) && (m_half_band == _other.m_half_band));
            if(!tmp) return false;
            
            const_pointer ptmp1 = m_start, ptmp2 = _other.m_start;
            while(ptmp1 != m_finish){
                tmp &= (*ptmp1 == *ptmp2);
                if(!tmp) return false;
                ++ptmp1;
                ++ptmp2;
            }
            return true;
        }
        
        bool operator!=(self_type const& _other) const
        {
            bool tmp = ((m_size == _other.m_size) && (m_half_band == _other.m_half_band));
            if(!tmp) return true;
            
            const_pointer ptmp1 = m_start, ptmp2 = _other.m_start;
            while(ptmp1 != m_finish){
                tmp &= (*ptmp1 == *ptmp2);
                if(!tmp) return true;
                ++ptmp1;
                ++ptmp2;
            }
            return false;
        }
        
        void clear()
        {
            size_type size1 = storage_size();
            for(size_type i1 = 0; i1 < size1; ++i1)
                m_allocator.destroy(m_start + i1);
            m_allocator.deallocate(m_start, size1);
            m_size = 0;
            m_half_band = 0;
            m_start = 0;
            m_finish = 0;
        }
        
        //---- iterators
        typedef std::ptrdiff_t difference_type;
        template<bool IsConst> class iterator1_;
        template<bool IsConst> class iterator2_;
        template<bool IsConst> class sup_iterator2_;
        template<bool IsConst> class sup_iterator1_;
        
        typedef iterator1_<false> iterator1;
        typedef iterator2_<false> iterator2;
        typedef iterator1_<true> const_iterator1;
        typedef iterator2_<true> const_iterator2;
        typedef sup_iterator2_<false> sup_iterator2;
        typedef sup_iterator1_<false> sup_iterator1;
        typedef sup_iterator2_<true> const_sup_iterator2;
        typedef sup_iterator1_<true> const_sup_iterator1;
        typedef reverse_iterator<iterator1> reverse_iterator1;
        typedef reverse_iterator<iterator2> reverse_iterator2;
        typedef reverse_iterator<const_iterator1> const_reverse_iterator1;
        typedef reverse_iterator<const_iterator1> const_reverse_iterator2;
        typedef reverse_iterator<sup_iterator2> reverse_sup_iterator2;
        typedef reverse_iterator<sup_iterator1> reverse_sup_iterator1;
        typedef reverse_iterator<const_sup_iterator2> const_reverse_sup_iterator2;
        typedef reverse_iterator<const_sup_iterator1> const_reverse_sup_iterator1;
        
        // begin1 end1
        iterator1 begin1()
        { return iterator1(*this, 0, 0); }
        const_iterator1 begin1() const
        { return const_iterator1(*this, 0, 0); }
        iterator1 end1()
        { return iterator1(*this, m_size, 0); }
        const_iterator1 end1() const
        { return const_iterator1(*this, m_size, 0); }
        
        //rbegin1 rend1
        reverse_iterator1 rbegin1()
        { return reverse_iterator1(end1()); }
        const_reverse_iterator1 rbegin1() const
        { return const_reverse_iterator(end1()); }
        reverse_iterator1 rend1()
        { return reverse_iterator1(begin1()); }
        const_reverse_iterator1 rend1() const
        { return const_reverse_iterator1(begin1()); }
        
        // begin2 end2
        iterator2 begin2()
        { return iterator2(*this, 0, 0); }
        const_iterator2 begin2() const
        { return const_iterator2(*this, 0, 0); }
        iterator2 end2()
        { return iterator2(*this, 0, m_size); }
        const_iterator2 end2() const
        { return const_iterator2(*this, 0, m_size); }
        
        //rbegin2 rend2
        reverse_iterator2 rbegin2()
        { return reverse_iterator2(end2()); }
        const_reverse_iterator2 rbegin2() const
        { return const_reverse_iterator(end2()); }
        reverse_iterator2 rend2()
        { return reverse_iterator2(begin2()); }
        const_reverse_iterator2 rend2() const
        { return const_reverse_iterator2(begin2()); }
        
    public:
        
        //-----------------------------------------------------------
        //    class iterator1_
        //-----------------------------------------------------------
        template<bool IsConst>
        class iterator1_
        {
        public:
            static bool const is_const_iterator = IsConst;
            typedef iterator1_ self_type;
            typedef sym_banded_matrix matrix_type;
            typedef sym_banded_matrix::value_type value_type;
            typedef std::ptrdiff_t difference_type;
            typedef typename
                std::conditional<IsConst, value_type const*, value_type*>::type pointer;
            typedef typename
                std::conditional<IsConst, value_type const&, value_type&>::type reference;
            typedef typename
                std::conditional<IsConst, matrix_type const*, matrix_type*>::type matrix_pointer;
            typedef typename
                std::conditional<IsConst, matrix_type const&, matrix_type&>::type matrix_reference;
            typedef std::random_access_iterator_tag iterator_category;
            
            typedef sup_iterator2_<IsConst> dual_iterator;
            typedef std::reverse_iterator<dual_iterator> dual_reverse_iterator;
            
            //---- constructor
            inline iterator1_() : m_row(0), m_col(0), m_mat(0) {}
            inline iterator1_(matrix_reference _mat, size_type _row, size_type _col)
                : m_row(_row), m_col(_col), m_mat(&_mat) {}
            
            inline iterator1_(iterator1_<false> const& _other)
                : m_row(_other.index1()), m_col(_other.index2()), m_mat(_other.base()) {}
            
            template<bool B>
            inline iterator1_(iterator2_<B> const& _other)
                : m_row(_other.index1()), m_col(_other.index2()), m_mat(_other.base()) {}
            
            //---- operators
            inline reference operator*() const
            {
                return (*m_mat)(m_row, m_col);
            }
            
            inline self_type& operator++()
            {
                ++m_row;
                return *this;
            }
        
            inline self_type operator++(int)
            {
                self_type tmp(*this);
                ++m_row;
                return tmp;
            }
        
            inline self_type& operator--()
            {
                --m_row;
                return *this;
            }
        
            inline self_type operator--(int)
            {
                self_type tmp(*this);
                --m_row;
                return tmp;
            }
            
            template<bool B>
            inline difference_type operator-(iterator1_<B> const& _other) const
            {
                return difference_type(m_row - _other.index1());
            }
            
            template<bool B>
            inline bool operator==(iterator1_<B> const& _other) const
            {
                return (m_mat == _other.base()
                    && m_row == _other.index1()
                    && m_col == _other.index2());
            }
        
            template<bool B>
            inline bool operator!=(iterator1_<B> const& _other) const
            {
                return (m_row != _other.index1()
                    || m_col != _other.index2()
                    || m_mat != _other.base());
            }
            
            //---- index
            inline size_type index1() const
            { return m_row; }
            inline size_type index2() const
            { return m_col; }
            
            //---- begin end
            inline dual_iterator begin() const
            {
                size_type hband = m_mat->get_half_band_width();
                pointer start = m_mat->data();
                if(m_row < hband){
                    return dual_iterator(*m_mat, start + m_row, m_row, 0);
                }else{
                    return dual_iterator(*m_mat, start + hband * (m_row - hband + 2) - 1, m_row, m_row -  hband + 1);
                }
            }
            
            inline dual_iterator end() const
            {
                size_type hband = m_mat->get_half_band_width();
                pointer start = m_mat->data();
                return dual_iterator(*m_mat, start + hband * (m_row + 1) - 1, m_row, m_row + 1);
            }
            
            inline dual_reverse_iterator rbegin() const
            {
                return dual_reverse_iterator(end());
            }
            
            inline dual_reverse_iterator rend() const
            {
                return dual_reverse_iterator(begin());
            }
            
            matrix_pointer base() const
            { return m_mat; }
            
        private:
            size_type m_row;
            size_type m_col;
            matrix_pointer m_mat;
        };//---- class iterator1_
        
        
        //-----------------------------------------------------------
        //    class iterator2_
        //-----------------------------------------------------------
        template<bool IsConst>
        class iterator2_
        {
        public:
            static bool const is_const_iterator = IsConst;
            typedef iterator2_ self_type;
            typedef sym_banded_matrix matrix_type;
            typedef sym_banded_matrix::value_type value_type;
            typedef std::ptrdiff_t difference_type;
            typedef typename
                std::conditional<IsConst, value_type const*, value_type*>::type pointer;
            typedef typename
                std::conditional<IsConst, value_type const&, value_type&>::type reference;
            typedef typename
                std::conditional<IsConst, matrix_type const*, matrix_type*>::type matrix_pointer;
            typedef typename
                std::conditional<IsConst, matrix_type const&, matrix_type&>::type matrix_reference;
            typedef std::random_access_iterator_tag iterator_category;
            
            typedef sup_iterator1_<IsConst> dual_iterator;
            typedef std::reverse_iterator<dual_iterator> dual_reverse_iterator;
            
            //---- constructor
            inline iterator2_() : m_row(0), m_col(0), m_mat(0) {}
            inline iterator2_(matrix_reference _mat, size_type _row, size_type _col)
                : m_row(_row), m_col(_col), m_mat(&_mat) {}
            
            inline iterator2_(iterator2_<false> const& _other)
                : m_row(_other.index1()), m_col(_other.index2()), m_mat(_other.base()) {}
            
            template<bool B>
            inline iterator2_(iterator1_<B> const& _other)
                : m_row(_other.index1()), m_col(_other.index2()), m_mat(_other.base()) {}
            
            //---- operators
            inline reference operator*() const
            {
                return (*m_mat)(m_row, m_col);
            }
            
            inline self_type& operator++()
            {
                ++m_col;
                return *this;
            }
        
            inline self_type operator++(int)
            {
                self_type tmp(*this);
                ++m_col;
                return tmp;
            }
        
            inline self_type& operator--()
            {
                --m_col;
                return *this;
            }
        
            inline self_type operator--(int)
            {
                self_type tmp(*this);
                --m_col;
                return tmp;
            }
            
            template<bool B>
            inline difference_type operator-(iterator2_<B> const& _other) const
            {
                return difference_type(m_col - _other.index2());
            }
            
            template<bool B>
            inline bool operator==(iterator2_<B> const& _other) const
            {
                return (m_mat == _other.base()
                    && m_row == _other.index1()
                    && m_col == _other.index2());
            }
        
            template<bool B>
            inline bool operator!=(iterator2_<B> const& _other) const
            {
                return (m_row != _other.index1()
                    || m_col != _other.index2()
                    || m_mat != _other.base());
            }
            
            //---- index
            inline size_type index1() const
            { return m_row; }
            inline size_type index2() const
            { return m_col; }
            
            //---- begin end
            inline dual_iterator begin() const
            {
                size_type hband = m_mat->get_half_band_width();
                pointer start = m_mat->data();
                return dual_iterator(*m_mat, start + m_col * hband, m_col, m_col);
            }
            
            inline dual_iterator end() const
            {
                size_type hband = m_mat->get_half_band_width();
                size_type mat_size = m_mat->size();
                pointer start = m_mat->data();
                if(m_col <= (mat_size - hband)){
                    return dual_iterator(*m_mat, start + (m_col + 1) * hband, m_col + hband, m_col);
                }else{
                    return dual_iterator(*m_mat, start + m_col * hband + mat_size - m_col, m_mat->size(), m_col);
                }
            }
            
            inline dual_reverse_iterator rbegin() const
            { return dual_reverse_iterator(end()); }
            
            inline dual_reverse_iterator rend() const
            { return dual_reverse_iterator(begin()); }
            
            matrix_pointer base() const
            { return m_mat; }
            
        private:
            size_type m_row;
            size_type m_col;
            matrix_pointer m_mat;
        };//---- class iterator2_
        
        //-----------------------------------------------------------
        //    class sup_iterator2_
        //-----------------------------------------------------------
        template<bool IsConst>
        class sup_iterator2_
            : public interval_iterator<
                typename std::conditional<
                    IsConst, value_type const*, value_type*
                >::type, sym_banded_matrix>
        {
        public:
            static bool const is_const_iterator = IsConst;
            typedef sup_iterator2_ self_type;
            typedef sym_banded_matrix matrix_type;
            typedef interval_iterator<
                typename std::conditional<
                    IsConst, value_type const*, value_type*
                >::type, sym_banded_matrix> base_iterator_type;
            typedef typename
                std::conditional<IsConst, matrix_type const&,
                    matrix_type&>::type matrix_reference;
            typedef typename base_iterator_type::pointer pointer;
            
            typedef sup_iterator1_<IsConst> dual_iterator;
            typedef std::reverse_iterator<dual_iterator> dual_reverse_iterator;
            
            //---- constructor
            inline sup_iterator2_()
                : base_iterator_type(0, 0), m_size(0), m_half_band(0), m_row(0), m_col(0), m_start(0) {}
            
            inline sup_iterator2_(matrix_reference _mat, pointer _pos, size_type _row, size_type _col)
                : base_iterator_type(_mat.get_half_band_width() - 1, _pos),
                    m_size(_mat.size()), m_half_band(_mat.get_half_band_width()),
                    m_row(_row), m_col(_col), m_start(_mat.data()) {}
            
            inline sup_iterator2_(size_type _size, size_type _hband, pointer _start, pointer _pos, size_type _row, size_type _col)
                : base_iterator_type(_hband - 1, _pos),
                    m_size(_size), m_half_band(_hband),
                    m_row(_row), m_col(_col), m_start(_start) {}
            
            inline sup_iterator2_(sup_iterator2_<false> const& _other)
                : base_iterator_type(_other.get_interval(), _other.base()),
                    m_size(_other.matrix_size()), m_half_band(_other.get_half_band_width()),
                    m_row(_other.index1()), m_col(_other.index2()), m_start(_other.get_start()) {}
            
            template<bool B>
            inline explicit sup_iterator2_(sup_iterator1_<B> const& _other)
                : base_iterator_type(_other.get_half_band_width() - 1, _other.base()),
                    m_size(_other.matrix_size()), m_half_band(_other.get_half_band_width()),
                    m_row(_other.index1()), m_col(_other.index2()), m_start(_other.get_start()) {}
            //----
            inline self_type& operator++()
            {
                ++m_col;
                base_iterator_type::operator++();
                return (*this);
            }
    
            inline self_type operator++(int)
            {
                ++m_col;
                self_type tmp(*this);
                (*this).base_iterator_type::operator++();
                return tmp;
            }
    
            inline self_type& operator--()
            {
                --m_col;
                base_iterator_type::operator--();
                return (*this);
            }
    
            inline self_type operator--(int)
            {
                --m_col;
                self_type tmp(*this);
                tmp.base_iterator_type::operator--();
                return tmp;
            }
        
            //---- index
            inline size_type index1() const
            { return m_row; }
            
            inline size_type index2() const
            { return m_col; }
            
            pointer get_start() const
            { return m_start; }
            
            size_type get_half_band_width() const
            { return m_half_band; }
            
            size_type matrix_size() const
            { return m_size; }
            
            //---- begin end
            inline dual_iterator begin() const
            {
                pointer tmp = m_start + index2() * m_half_band;
                return dual_iterator(m_size, m_half_band, m_start, tmp, m_col, m_col);
            }
            
            inline dual_iterator end() const
            {
                if(m_col <= (m_size - m_half_band)){
                    pointer tmp = m_start + (m_col + 1) * m_half_band;
                    return dual_iterator(m_size, m_half_band, m_start, tmp, m_col + m_half_band, m_col);
                }else{
                    pointer tmp = m_start + m_col * m_half_band + m_size - m_col;
                    return dual_iterator(m_size, m_half_band, m_start, tmp, m_size, m_col);
                }
            }
            
            inline dual_reverse_iterator rbegin() const
            {
                return dual_reverse_iterator(end());
            }
            
            inline dual_reverse_iterator rend() const
            {
                return dual_reverse_iterator(begin());
            }
            
        private:
            size_type m_size;
            size_type m_half_band;
            size_type m_row;
            size_type m_col;
            pointer m_start;
        };//---- class sup_iterator2_
        
        
        //-----------------------------------------------------------
        //    class sup_iterator1_
        //-----------------------------------------------------------
        template<bool IsConst>
        class sup_iterator1_
            : public interval_iterator<
                typename std::conditional<
                    IsConst, value_type const*, value_type*
                >::type, sym_banded_matrix>
        {
        public:
            static bool const is_const_iterator = IsConst;
            typedef sup_iterator1_ self_type;
            typedef sym_banded_matrix matrix_type;
            typedef interval_iterator<
                typename std::conditional<
                    IsConst, value_type const*, value_type*
                >::type, sym_banded_matrix> base_iterator_type;
            typedef typename
                std::conditional<IsConst, matrix_type const&,
                    matrix_type&>::type matrix_reference;
            typedef typename base_iterator_type::pointer pointer;
            
            typedef sup_iterator2_<IsConst> dual_iterator;
            typedef std::reverse_iterator<dual_iterator> dual_reverse_iterator;
            
            //---- constructor
            inline sup_iterator1_()
                : base_iterator_type(1, 0), m_half_band(0), m_row(0), m_col(0), m_start(0) {}
            
            inline sup_iterator1_(matrix_reference _mat, pointer _pos, size_type _row, size_type _col)
                : base_iterator_type(1, _pos),
                    m_size(_mat.size()), m_half_band(_mat.get_half_band_width()),
                    m_row(_row), m_col(_col), m_start(_mat.data()) {}
            
            inline sup_iterator1_(size_type _size, size_type _hband, pointer _start, pointer _pos, size_type _row, size_type _col)
                : base_iterator_type(1, _pos),
                    m_size(_size), m_half_band(_hband),
                    m_row(_row), m_col(_col), m_start(_start) {}
            
            inline sup_iterator1_(sup_iterator1_<false> const& _other)
                : base_iterator_type(1, _other.base()),
                    m_size(_other.matrix_size()), m_half_band(_other.get_half_band_width()),
                    m_row(_other.index1()), m_col(_other.index2()), m_start(_other.get_start()) {}
            
            template<bool B>
            inline explicit sup_iterator1_(sup_iterator2_<B> const& _other)
                : base_iterator_type(1, _other.base()),
                    m_size(_other.matrix_size()), m_half_band(_other.get_half_band_width()),
                    m_row(_other.index1()), m_col(_other.index2()), m_start(_other.get_start()) {}
            //----
            inline self_type& operator++()
            {
                ++m_row;
                base_iterator_type::operator++();
                return (*this);
            }
    
            inline self_type operator++(int)
            {
                ++m_row;
                self_type tmp(*this);
                (*this).base_iterator_type::operator++();
                return tmp;
            }
    
            inline self_type& operator--()
            {
                --m_row;
                base_iterator_type::operator--();
                return (*this);
            }
    
            inline self_type operator--(int)
            {
                --m_row;
                self_type tmp(*this);
                tmp.base_iterator_type::operator--();
                return tmp;
            }
            
            //---- index
            inline size_type index1() const
            { return m_row; }
            
            inline size_type index2() const
            { return m_col; }
            
            pointer get_start() const
            { return m_start; }
            
            size_type get_half_band_width() const
            { return m_half_band; }
            
            size_type matrix_size() const
            { return m_size; }
            
            //---- begin end
            inline dual_iterator begin() const
            {
                if(m_row < m_half_band){
                    return dual_iterator(m_size, m_half_band, m_start, m_start + m_row, m_row, 0);
                }else{
                    pointer tmp = m_start + m_half_band * (m_row - m_half_band + 2) - 1;
                    return dual_iterator(m_size, m_half_band, m_start, tmp, m_row, m_row - m_half_band + 1);
                }
            }
            
            inline dual_iterator end() const
            {
                pointer tmp = m_start + m_half_band * (m_row + 1) - 1;
                return dual_iterator(m_size, m_half_band, m_start, tmp, m_row, m_row + 1);
            }
            
            inline dual_reverse_iterator rbegin() const
            {
                return dual_reverse_iterator(end());
            }
            
            inline dual_reverse_iterator rend() const
            {
                return dual_reverse_iterator(begin());
            }
            
        private:
            size_type m_size;
            size_type m_half_band;
            size_type m_row;
            size_type m_col;
            pointer m_start;
        };//---- class sup_iterator1_
    };
    
    //---- sym_banded_matrix m_zero
    template<typename Tp, typename Allocator>
    typename sym_banded_matrix<Tp, Allocator>::value_type const
        sym_banded_matrix<Tp, Allocator>::m_zero = sym_banded_matrix<Tp, Allocator>::value_type();
}}//---- namespace numeric sykes


//-----------------------------------------------------------
//    prod m v
//-----------------------------------------------------------
template<typename Tp, typename MatAlloc, typename VecAlloc>
sykes::numeric::vector<Tp, VecAlloc>
prod(sykes::numeric::sym_banded_matrix<Tp, MatAlloc> const& _mat, sykes::numeric::vector<Tp, VecAlloc> const& _vec)
{
    typedef typename sykes::numeric::sym_banded_matrix<Tp>::size_type size_type;
    size_type msize = _mat.size(), vsize = _vec.size();
    assert(msize == vsize);
    if(msize != vsize){
        throw std::runtime_error("prod(_mat, _vec)");
    }
    sykes::numeric::vector<Tp> vtmp(msize, Tp());
    
    auto it1 = _mat.begin1(), end1 = _mat.end1();
    auto vit = vtmp.begin();
    while(it1 != end1){
        auto it2 = it1.begin(), end2 = it1.end();
        --end2;
        while(it2 != end2){
            size_type pos = it2.index2();
            *vit += (*it2) * _vec(pos);
            ++it2;
        }
        
        typename sykes::numeric::sym_banded_matrix<Tp, MatAlloc>::const_sup_iterator1
            it3(it2), end3 = it2.end();
        
        while(it3 != end3){
            size_type pos = it3.index1();
            *vit += (*it3) * _vec(pos);
            ++it3;
        }
        ++it1;
        ++vit;
    }
    return vtmp;
}

//-----------------------------------------------------------
//    prod v m
//-----------------------------------------------------------
template<typename Tp, typename MatAlloc, typename VecAlloc>
sykes::numeric::vector<Tp, VecAlloc>
prod(sykes::numeric::vector<Tp, VecAlloc> const& _vec, sykes::numeric::sym_banded_matrix<Tp, MatAlloc> const& _mat)
{
    typedef typename sykes::numeric::sym_banded_matrix<Tp>::size_type size_type;
    size_type msize = _mat.size(), vsize = _vec.size();
    assert(msize == vsize);
    if(msize != vsize){
        throw std::runtime_error("prod(_mat, _vec)");
    }
    sykes::numeric::vector<Tp> vtmp(msize, Tp());
    
    auto it1 = _mat.begin1(), end1 = _mat.end1();
    auto vit = vtmp.begin();
    while(it1 != end1){
        auto it2 = it1.begin(), end2 = it1.end();
        --end2;
        while(it2 != end2){
            size_type pos = it2.index2();
            *vit += (*it2) * _vec(pos);
            ++it2;
        }
        
        typename sykes::numeric::sym_banded_matrix<Tp, MatAlloc>::const_sup_iterator1
            it3(it2), end3 = it2.end();
        
        while(it3 != end3){
            size_type pos = it3.index1();
            *vit += (*it3) * _vec(pos);
            ++it3;
        }
        ++it1;
        ++vit;
    }
    return vtmp;
}
#endif


