//-----------------------------------------------------------
//    vector.h
//-----------------------------------------------------------
#ifndef SYKES_NUMERIC_VECTOR_H
#define SYKES_NUMERIC_VECTOR_H

#include <iostream>
#include <stdexcept>
#include <memory>
#include <iterator>
#include <type_traits>
#include <cstddef>
#include <assert.h>
#include <new>
#include "matrix_and_vector_base.h"
#include "wrapping_iterator.h"

namespace sykes{ namespace numeric{
    //-----------------------------------------------------------
    //    class sym_banded_matrix
    //-----------------------------------------------------------
    template<typename Tp, typename Allocator = std::allocator<Tp> >
    class vector : public vector_base
    {
    public:
        typedef vector self_type;
        typedef Tp value_type;
        typedef Allocator allocator_type;
        typedef typename Allocator::reference reference;
        typedef typename Allocator::const_reference const_reference;
        typedef typename Allocator::size_type size_type;
        typedef typename Allocator::pointer pointer;
        typedef typename Allocator::const_pointer const_pointer;
    
    private:
        Allocator m_allocator;
        pointer m_start;
        pointer m_finish;
        
    public:
        
        //---- constructor
        vector()
            : m_allocator(), m_start(0), m_finish(0)
        {}
        
        vector(size_type _size, value_type const& _val = value_type())
            : m_allocator(), m_start(0), m_finish(0)
        {
            m_start = m_allocator.allocate(_size);
            m_finish = m_start + _size;
            std::uninitialized_fill(m_start, m_finish, _val);
        }
        
        //---- copy constructor
        vector(self_type const& _other)
            : m_allocator(), m_start(0), m_finish(0)
        {
            m_start = m_allocator.allocate(_other.size());
            m_finish = std::uninitialized_copy(_other.m_start, _other.m_finish, m_start);
        }
        
        //---- destructor
        ~vector()
        {
            size_type sz = size();
            for(size_type i1 = 0; i1 < sz; ++i1)
                m_allocator.destroy(m_start + i1);
            m_allocator.deallocate(m_start, sz);
        }
        
        //---- resize
        void resize(size_type _size, bool _preserve = true, value_type const& _val = value_type())
        {
            size_type new_size = _size;
            size_type old_size = size();
            
            if(_size == old_size){
                if(!_preserve) std::fill(m_start, m_finish, _val);
                return;
            }else if(new_size > max_size())
                throw std::length_error("sykes::numeric::vector::resize()");
            else{
                pointer tmp = m_allocator.allocate(new_size);
                if(_preserve){
                    if(new_size > old_size){
                        std::uninitialized_copy(m_start, m_finish, tmp);
                        std::uninitialized_fill(tmp + old_size, tmp + new_size, _val);
                    }else{
                        std::uninitialized_copy(m_start, m_start + new_size, tmp);
                    }
                }else{
                    std::uninitialized_fill(tmp, tmp + new_size, _val);
                }
                
                for(size_type i1 = 0; i1 < old_size; ++i1)
                    m_allocator.destroy(m_start + i1);
                m_allocator.deallocate(m_start, old_size);
                
                m_start = tmp;
                m_finish = tmp + new_size;
            }
        }
        
        //---- operator=
        self_type& operator=(self_type const& _other)
        {
            size_type new_size = _other.size();
            size_type old_size = size();
            if(new_size > max_size())
                throw std::length_error("sykes::numeric::vector::resize()");
            else if(old_size == new_size){
                std::copy(_other.m_start, _other.m_finish, m_start);
                return (*this);
            }else{
                pointer tmp = m_allocator.allocate(new_size);
                std::uninitialized_copy(_other.m_start, _other.m_finish, tmp);
                
                for(size_type i1 = 0; i1 < old_size; ++i1)
                    m_allocator.destroy(m_start + i1);
                m_allocator.deallocate(m_start, old_size);
                
                m_start = tmp;
                m_finish = tmp + new_size;
                return (*this);
            }
        }
        
        //---- element access
        reference operator()(size_type _x)
        {
            assert(_x < size());
            return *(m_start + _x);
        }
        
        const_reference operator()(size_type _x) const
        {
            assert(_x < size());
            return *(m_start + _x);
        }
        
        value_type& at(size_type _x)
        {
            assert(_x < size());
            if(_x >= size()){
                throw std::range_error("sykes::numeric::vector::at()");
            }
            return *(m_start + _x);
        }
        
        value_type const& at(size_type _x) const
        {
            assert(_x < size());
            if(_x >= size()){
                throw std::range_error("sykes::numeric::vector::at()");
            }
            return *(m_start + _x);
        }
        
        //---- front back
        const_reference front() const { return *m_start; }
        reference front() { return *m_start; }
        const_reference back() const { return *(m_finish - 1); }
        reference back() { return *(m_finish - 1); }
        
        //---- size
        size_type size() const { return size_type(m_finish - m_start); }
        size_type max_size() const { return m_allocator.max_size(); }
        bool empty() const { return (m_start == m_finish); }
        
        //---- data()
        Tp const* data() const { return m_start; }
        Tp* data() { return m_start; }
        
        //---- get_allocator
        allocator_type get_allocator() const { return m_allocator; }
        
        //---- comparison
        bool operator==(self_type const& _other) const
        {
            bool tmp = (size() == _other.size());
            if(!tmp) return false;
            const_pointer ptmp1 = m_start, ptmp2 = _other.m_start;
            while(ptmp1 != m_finish){
                tmp &= (*ptmp1 == *ptmp2);
                if(!tmp) return false;
                ++ptmp1;
                ++ptmp2;
            }
            return true;
        }
        
        bool operator!=(self_type const& _other) const
        {
            bool tmp = (size() == _other.size());
            if(!tmp) return true;
            const_pointer ptmp1 = m_start, ptmp2 = _other.m_start;
            while(ptmp1 != m_finish){
                tmp &= (*ptmp1 == *ptmp2);
                if(!tmp) return true;
                ++ptmp1;
                ++ptmp2;
            }
            return false;
        }
        
        //---- iterators
        typedef std::ptrdiff_t difference_type;
        typedef wrapping_iterator<pointer, self_type> iterator;
        typedef wrapping_iterator<const_pointer, self_type> const_iterator;
        typedef std::reverse_iterator<iterator> reverse_iterator;
        typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
        
        // begin end
        iterator begin()
        { return iterator(m_start); }
        const_iterator begin() const
        { return const_iterator(m_start); }
        iterator end()
        { return iterator(m_finish); }
        const_iterator end() const
        { return const_iterator(m_finish); }
        
        //rbegin rend
        reverse_iterator rbegin()
        { return reverse_iterator(end()); }
        const_reverse_iterator rbegin() const
        { return const_reverse_iterator(end()); }
        reverse_iterator rend()
        { return reverse_iterator(begin()); }
        const_reverse_iterator rend() const
        { return const_reverse_iterator(begin()); }
        
        //---- operator
        self_type& operator+=(self_type const& _vec)
        {
            size_type vsize1 = size(), vsize2 = _vec.size();
            assert(vsize1 == vsize2);
            if(vsize1 != vsize2){
                throw std::runtime_error("sykes::numeric::vector::operator+=()");
            }
            auto it1 = begin(), end1 = end();
            auto it2 = _vec.begin();
            while(it1 != end1){
                (*it1) += (*it2);
                ++it1;
                ++it2;
            }
            return (*this);
        }
        
        self_type& operator-=(self_type const& _vec)
        {
            size_type vsize1 = size(), vsize2 = _vec.size();
            assert(vsize1 == vsize2);
            if(vsize1 != vsize2){
                throw std::runtime_error("sykes::numeric::vector::operator-=()");
            }
            auto it1 = begin(), end1 = end();
            auto it2 = _vec.begin();
            while(it1 != end1){
                (*it1) -= (*it2);
                ++it1;
                ++it2;
            }
            return (*this);
        }
        
        self_type& operator*=(value_type const& _x)
        {
            auto it1 = begin(), end1 = end();
            while(it1 != end1){
                (*it1) *= _x;
                ++it1;
            }
            return (*this);
        }
        
        self_type& operator/=(value_type const& _x)
        {
            auto it1 = begin(), end1 = end();
            while(it1 != end1){
                (*it1) /= _x;
                ++it1;
            }
            return (*this);
        }
    };
    
    template<typename Tp, typename Allocator>
    vector<Tp, Allocator> operator*(Tp const& _x, vector<Tp, Allocator> const& _vec)
    {
        vector<Tp, Allocator> vtmp(_vec);
        auto it1 = vtmp.begin(), end1 = vtmp.end();
        while(it1 != end1){
            *it1 *= _x;
            ++it1;
        }
        return vtmp;
    }
    
    template<typename Tp, typename Allocator>
    vector<Tp, Allocator> operator*(vector<Tp, Allocator> const& _vec, Tp const& _x)
    {
        vector<Tp, Allocator> vtmp(_vec);
        auto it1 = vtmp.begin(), end1 = vtmp.end();
        while(it1 != end1){
            *it1 *= _x;
            ++it1;
        }
        return vtmp;
    }
    
    template<typename Tp, typename Allocator>
    vector<Tp, Allocator> operator/(vector<Tp, Allocator> const& _vec, Tp const& _x)
    {
        vector<Tp, Allocator> vtmp(_vec);
        auto it1 = vtmp.begin(), end1 = vtmp.end();
        while(it1 != end1){
            *it1 /= _x;
            ++it1;
        }
        return vtmp;
    }
    
    template<typename Tp, typename Allocator>
    vector<Tp, Allocator> operator+(vector<Tp, Allocator> const& _vec1, vector<Tp, Allocator> const& _vec2)
    {
        typedef typename vector<Tp, Allocator>::size_type size_type;
        size_type vsize1 = _vec1.size(), vsize2 = _vec2.size();
        assert(vsize1 == vsize2);
        if(vsize1 != vsize2){
            throw std::runtime_error("inner_prod(_vec, _vec)");
        }
        vector<Tp, Allocator> vtmp(vsize1, Tp());
        
        size_type n1 = 0;
        auto it1 = _vec1.begin(), end1 = _vec1.end();
        auto it2 = _vec2.begin();
        
        while(n1 != vsize1){
            vtmp(n1) = (*it1) + (*it2);
            ++it1;
            ++it2;
            ++n1;
        }
        return vtmp;
    }
    
    template<typename Tp, typename Allocator>
    vector<Tp, Allocator> operator-(vector<Tp, Allocator> const& _vec1, vector<Tp, Allocator> const& _vec2)
    {
        typedef typename vector<Tp, Allocator>::size_type size_type;
        size_type vsize1 = _vec1.size(), vsize2 = _vec2.size();
        assert(vsize1 == vsize2);
        if(vsize1 != vsize2){
            throw std::runtime_error("inner_prod(_vec, _vec)");
        }
        vector<Tp, Allocator> vtmp(vsize1, Tp());
        
        size_type n1 = 0;
        auto it1 = _vec1.begin(), end1 = _vec1.end();
        auto it2 = _vec2.begin();
        
        while(n1 != vsize1){
            vtmp(n1) = (*it1) - (*it2);
            ++it1;
            ++it2;
            ++n1;
        }
        return vtmp;
    }
}}//---- namespace numeric sykes

//-----------------------------------------------------------
//    inner_prod
//-----------------------------------------------------------
template<typename Tp, typename Alloc1, typename Alloc2>
Tp inner_prod(sykes::numeric::vector<Tp, Alloc1> const& _vec1, sykes::numeric::vector<Tp, Alloc2> const& _vec2)
{
    typedef typename sykes::numeric::vector<Tp>::size_type size_type;
    size_type vsize1 = _vec1.size(), vsize2 = _vec2.size();
    assert(vsize1 == vsize2);
    if(vsize1 != vsize2){
        throw std::runtime_error("inner_prod()");
    }
    
    auto it1 = _vec1.begin(), end1 = _vec1.end();
    auto it2 = _vec2.begin();
    
    Tp tmp = Tp();
    while(it1 != end1){
        tmp += (*it1) * (*it2);
        ++it1;
        ++it2;
    }
    return tmp;
}

#endif


