//-----------------------------------------------------------
//    wrapping_iterator.h
//-----------------------------------------------------------
#ifndef SYKES_NUMERIC_WRAPPING_ITERATOR
#define SYKES_NUMERIC_WRAPPING_ITERATOR
#include <stdexcept>
#include <memory>
#include <iterator>
#include <type_traits>
#include <cstddef>
#include <assert.h>

namespace sykes{ namespace numeric{
    //-----------------------------------------------------------
    //    class wrapping_iterator
    //-----------------------------------------------------------
    template<typename _Iterator, typename _Container>
    class wrapping_iterator
    {
    public:
        typedef wrapping_iterator self_type;
        typedef typename std::iterator_traits<_Iterator>::value_type
            value_type;
        typedef typename std::iterator_traits<_Iterator>::difference_type
            difference_type;
        typedef typename std::iterator_traits<_Iterator>::pointer
            pointer;
        typedef typename std::iterator_traits<_Iterator>::reference
            reference;
        typedef typename std::iterator_traits<_Iterator>::iterator_category
            iterator_category;
        
        //---- constructor
        wrapping_iterator() : m_current() {}
        explicit wrapping_iterator(_Iterator const& ite0) : m_current(ite0) {}
        
        //---- construct const iterator from iterator
        template<typename _Ite>
        wrapping_iterator(
            wrapping_iterator<
                _Ite,
                typename std::enable_if<
                    (std::is_same<
                        typename std::iterator_traits<_Ite>::pointer,
                        typename _Container::pointer
                    >::value), _Container
                >::type
            > const& other)
        : m_current(other.base()) {}
        
        reference operator*() const
        { return *m_current; }
        
        self_type& operator++()
        {
            ++m_current;
            return *this;
        }
        
        self_type operator++(int)
        { return self_type(m_current++); }
        
        self_type& operator--()
        {
            --m_current;
            return *this;
        }
        
        self_type operator--(int)
        { return self_type(m_current--); }
        
        //---- random access
        reference operator[](difference_type const& n)
        {
            m_current += n;
            return *this;
        }
        
        self_type& operator+=(difference_type const& n)
        {
            m_current += n;
            return *this;
        }
        
        self_type operator+(difference_type const& n)
        { return self_type(m_current + n); }
        
        self_type operator-(difference_type const& n)
        { return self_type(m_current - n); }
        
        _Iterator const& base() const
        { return m_current; }
        
        //---- comparison
        bool operator==(self_type const& other) const
        { return (m_current == other.m_current); }
        
        bool operator!=(self_type const& other) const
        { return (base() != other.base()); }
        
        bool operator<(self_type const& other) const
        { return (base() < other.base()); }
        
        bool operator>(self_type const& other) const
        { return (base() > other.base()); }
        
        bool operator<=(self_type const& other) const
        { return (base() <= other.base()); }
        
        bool operator>=(self_type const& other) const
        { return (base() >= other.base()); }
        
        //---- comparison between iterator and const iterator
        template<typename _Ite>
        bool operator==(
            wrapping_iterator<
                _Ite,
                typename std::enable_if<
                    (std::is_same<
                        typename std::iterator_traits<_Ite>::pointer,
                        typename _Container::const_pointer
                    >::value), _Container
                >::type
            > const& other) const
        { return (base() == other.base()); }
        
        template<typename _Ite>
        bool operator!=(
            wrapping_iterator<
                _Ite,
                typename std::enable_if<
                    (std::is_same<
                        typename std::iterator_traits<_Ite>::pointer,
                        typename _Container::const_pointer
                    >::value), _Container
                >::type
            > const& other) const
        { return (base() != other.base()); }
        
        template<typename _Ite>
        bool operator<(
            wrapping_iterator<
                _Ite,
                typename std::enable_if<
                    (std::is_same<
                        typename std::iterator_traits<_Ite>::pointer,
                        typename _Container::const_pointer
                    >::value), _Container
                >::type
            > const& other) const
        { return (base() < other.base()); }
        
        template<typename _Ite>
        bool operator>(
            wrapping_iterator<
                _Ite,
                typename std::enable_if<
                    (std::is_same<
                        typename std::iterator_traits<_Ite>::pointer,
                        typename _Container::const_pointer
                    >::value), _Container
                >::type
            > const& other) const
        { return (base() > other.base()); }
        
        template<typename _Ite>
        bool operator<=(
            wrapping_iterator<
                _Ite,
                typename std::enable_if<
                    (std::is_same<
                        typename std::iterator_traits<_Ite>::pointer,
                        typename _Container::const_pointer
                    >::value), _Container
                >::type
            > const& other) const
        { return (base() <= other.base()); }
        
        template<typename _Ite>
        bool operator>=(
            wrapping_iterator<
                _Ite,
                typename std::enable_if<
                    (std::is_same<
                        typename std::iterator_traits<_Ite>::pointer,
                        typename _Container::const_pointer
                    >::value), _Container
                >::type
            > const& other) const
        { return (base() >= other.base()); }
    private:
        _Iterator m_current;
    };
    
    //---- iterator difference
    template<typename _Iterator, typename _Container>
    inline typename wrapping_iterator<_Iterator, _Container>::difference_type 
    operator-(wrapping_iterator<_Iterator, _Container> const& _lhs,
        wrapping_iterator<_Iterator, _Container> const& _rhs)
    { return _lhs.base() - _rhs.base(); }
    
    template<typename _LI, typename _RI, typename _Container>
    inline typename wrapping_iterator<_LI, _Container>::difference_type 
    operator-(
        wrapping_iterator<_LI, _Container> const& _lhs,
        wrapping_iterator<
            _RI,
            typename std::enable_if<
                (std::is_same<
                    typename std::iterator_traits<_RI>::pointer,
                    typename _Container::const_pointer
                >::value), _Container
            >::type
        > const& _rhs)
    { return _lhs.base() - _rhs.base(); }
}}//---- namespace numeric sykes
#endif

