#define BOOST_TEST_MODULE test_sym_banded_matrix
#include "sym_banded_matrix.h"
#include "matrix_and_vector_output.h"
#include <boost/test/included/unit_test.hpp>
#include <string>
#include <algorithm>
using namespace sykes::numeric;

BOOST_AUTO_TEST_CASE( construction )
{
	int size1 = 10, hbw = 3;
	sym_banded_matrix<double> m1(size1, hbw, 0);
	
	std::cout << m1 << std::endl;
	
	for(int i1 = 0; i1 < size1; ++i1){
		for(int i2 = 0; i2 <= i1; ++i2){
			if((i1 - i2) < hbw){
				m1(i1, i2) = i1 * 3 + i2;
			}
		}
	}
	
	std::cout << m1 << std::endl;
	
	sym_banded_matrix<double> m2(m1), m3(size1, hbw, 99);
	
	m3 = m2;
	
	std::cout << m2 << std::endl;
}

BOOST_AUTO_TEST_CASE( resize )
{
	int size1 = 10, hbw = 3;
	sym_banded_matrix<double> m1(size1, hbw, 0);
	for(int i1 = 0; i1 < size1; ++i1){
		for(int i2 = 0; i2 <= i1; ++i2){
			if((i1 - i2) < hbw){
				m1(i1, i2) = i1 * 3 + i2;
			}
		}
	}
	std::cout << m1 << std::endl;
	
	m1.resize(20, 6);
	
	BOOST_CHECK_EQUAL(m1.size(), 20);
	BOOST_CHECK_EQUAL(m1.get_half_band_width(), 6);
	BOOST_CHECK_EQUAL(m1.storage_size(), 20 * 6);
	
	std::cout << "resized" << std::endl;
	std::cout << m1 << std::endl;
	
	m1.resize(5, 2);
	
	BOOST_CHECK_EQUAL(m1.size(), 5);
	BOOST_CHECK_EQUAL(m1.get_half_band_width(), 2);
	BOOST_CHECK_EQUAL(m1.storage_size(), 5 * 2);
	
	std::cout << "re resized" << std::endl;
	std::cout << m1 << std::endl;
	
	m1.resize(10, 10, false, 9.9);
	BOOST_CHECK_EQUAL(m1.size(), 10);
	BOOST_CHECK_EQUAL(m1.get_half_band_width(), 10);
	BOOST_CHECK_EQUAL(m1.storage_size(), 10 * 10);
	std::cout << "re re resized" << std::endl;
	std::cout << m1 << std::endl;
	
}


struct Element
{
	static int construction_count;
	static int destruction_count;
	
	Element(){++construction_count;}
	Element(Element const& _other){++construction_count;}
	~Element(){++destruction_count;}
};
int Element::construction_count = 0;
int Element::destruction_count = 0;

BOOST_AUTO_TEST_CASE( element_construction_and_destruction_count )
{
	{
		int size1 = 100, band1 = 20;
		sym_banded_matrix<Element> X(size1, band1);
		sym_banded_matrix<Element> Y(X), Z;
		Z = Y;
	}
	std::cout << "construction count: " << Element::construction_count << std::endl;
	std::cout << "destruction count: " << Element::destruction_count << std::endl;
	BOOST_CHECK_EQUAL(Element::construction_count, Element::destruction_count);
	{
		int size1 = 10000, band1 = 20;
		sym_banded_matrix<Element> X(size1, band1);
	}
	std::cout << "construction count: " << Element::construction_count << std::endl;
	std::cout << "destruction count: " << Element::destruction_count << std::endl;
	BOOST_CHECK_EQUAL(Element::construction_count, Element::destruction_count);
	
}

BOOST_AUTO_TEST_CASE( size_test )
{
	const int size1 = 10, hbw = 3;
	sym_banded_matrix<double> m1(size1, hbw, 0);
	for(int i1 = 0; i1 < size1; ++i1){
		for(int i2 = 0; i2 <= i1; ++i2){
			if((i1 - i2) < hbw){
				m1(i1, i2) = i1 * 3 + i2;
			}
		}
	}
	BOOST_CHECK_EQUAL(m1.size(), size1);
	BOOST_CHECK_EQUAL(m1.size1(), size1);
	BOOST_CHECK_EQUAL(m1.size2(), size1);
	BOOST_CHECK_EQUAL(m1.get_half_band_width(), hbw);
	BOOST_CHECK_EQUAL(m1.storage_size(), size1 * hbw);
	
	sym_banded_matrix<double> m2;
	BOOST_CHECK_EQUAL(m2.size(), 0);
	
	m2 = m1;
	BOOST_CHECK_EQUAL(m2.size(), m1.size());
	BOOST_CHECK_EQUAL(m2.get_half_band_width(), m1.get_half_band_width());
	
	m2.clear();
	BOOST_CHECK_EQUAL(m2.size(), 0);
}

BOOST_AUTO_TEST_CASE( at_test )
{
	const int size1 = 10, hbw = 3;
	sym_banded_matrix<double> m1(size1, hbw, 0);
	for(int i1 = 0; i1 < size1; ++i1){
		for(int i2 = 0; i2 <= i1; ++i2){
			if((i1 - i2) < hbw){
				m1(i1, i2) = i1 * 3 + i2;
			}
		}
	}
	
	BOOST_CHECK_EQUAL(m1.at(5, 6), (6 * 3 + 5));
	BOOST_CHECK_THROW(m1.at(10, 10), std::exception );
	BOOST_CHECK_THROW(m1.at(9, 6), std::exception );
}

BOOST_AUTO_TEST_CASE( iterator1_test )
{
	const int size1 = 10, hbw = 3;
	sym_banded_matrix<int> m1(size1, hbw, 0);
	for(int i1 = 0; i1 < size1; ++i1){
		for(int i2 = 0; i2 <= i1; ++i2){
			if((i1 - i2) < hbw){
				m1(i1, i2) = i1 + 1;
			}
		}
	}
	std::cout << m1 << std::endl;
	{
		auto it1 = m1.begin1(), end1 = m1.end1();
		while(it1 != end1){
			auto it2 = it1.begin(), end2 = it1.end();
			int tmp = *it2;
			while(it2 != end2){
				BOOST_CHECK_EQUAL(tmp, *it2);
				std::cout << *it2 << ", ";
				++it2;
			}
			std::cout << std::endl;
			++it1;
		}
	}
	
	{
		auto it1 = m1.begin1(), end1 = m1.end1();
		while(it1 != end1){
			auto it2 = it1.rbegin(), end2 = it1.rend();
			int tmp = *it2;
			while(it2 != end2){
				BOOST_CHECK_EQUAL(tmp, *it2);
				std::cout << *it2 << ", ";
				++it2;
			}
			std::cout << std::endl;
			++it1;
		}
	}
	
	{
		auto it1 = m1.rbegin1(), end1 = m1.rend1();
		while(it1 != end1){
			auto it2 = it1.rbegin(), end2 = it1.rend();
			int tmp = *it2;
			while(it2 != end2){
				BOOST_CHECK_EQUAL(tmp, *it2);
				std::cout << *it2 << ", ";
				++it2;
			}
			std::cout << std::endl;
			++it1;
		}
	}
	
	{
		auto it1 = m1.rbegin1(), end1 = m1.rend1();
		while(it1 != end1){
			auto it2 = it1.rbegin(), end2 = it1.rend();
			int tmp = *it2;
			while(it2 != end2){
				BOOST_CHECK_EQUAL(tmp, *it2);
				std::cout << *it2 << ", ";
				++it2;
			}
			std::cout << std::endl;
			++it1;
		}
	}
	
}

BOOST_AUTO_TEST_CASE( iterator2_test )
{
	const int size1 = 10, hbw = 3;
	sym_banded_matrix<int> m1(size1, hbw, 0);
	for(int i1 = 0; i1 < size1; ++i1){
		for(int i2 = 0; i2 <= i1; ++i2){
			if((i1 - i2) < hbw){
				m1(i1, i2) = i2 + 1;
			}
		}
	}
	std::cout << m1 << std::endl;
	{
		auto it1 = m1.begin2(), end1 = m1.end2();
		while(it1 != end1){
			auto it2 = it1.begin(), end2 = it1.end();
			int tmp = *it2;
			while(it2 != end2){
				BOOST_CHECK_EQUAL(tmp, *it2);
				std::cout << *it2 << ", ";
				++it2;
			}
			std::cout << std::endl;
			++it1;
		}
	}
	
	{
		auto it1 = m1.begin2(), end1 = m1.end2();
		while(it1 != end1){
			auto it2 = it1.rbegin(), end2 = it1.rend();
			int tmp = *it2;
			while(it2 != end2){
				BOOST_CHECK_EQUAL(tmp, *it2);
				std::cout << *it2 << ", ";
				++it2;
			}
			std::cout << std::endl;
			++it1;
		}
	}
	
	{
		auto it1 = m1.rbegin2(), end1 = m1.rend2();
		while(it1 != end1){
			auto it2 = it1.rbegin(), end2 = it1.rend();
			int tmp = *it2;
			while(it2 != end2){
				BOOST_CHECK_EQUAL(tmp, *it2);
				std::cout << *it2 << ", ";
				++it2;
			}
			std::cout << std::endl;
			++it1;
		}
	}
	
	{
		auto it1 = m1.rbegin2(), end1 = m1.rend2();
		while(it1 != end1){
			auto it2 = it1.rbegin(), end2 = it1.rend();
			int tmp = *it2;
			while(it2 != end2){
				BOOST_CHECK_EQUAL(tmp, *it2);
				std::cout << *it2 << ", ";
				++it2;
			}
			std::cout << std::endl;
			++it1;
		}
	}
}

BOOST_AUTO_TEST_CASE( index_test )
{
	std::cout << "---------- index test ----------" << std::endl;
	const int size1 = 10, hbw = 3;
	sym_banded_matrix<int> m1(size1, hbw, 0);
	
	auto it2 = m1.begin2();
	auto it1 = it2.begin();
	
	std::cout << it1.index1() << ", " << it1.index2() << std::endl;
	BOOST_CHECK_EQUAL(it1.index1(), 0);
	BOOST_CHECK_EQUAL(it1.index2(), 0);
	++it1;
	
	std::cout << it1.index1() << ", " << it1.index2() << std::endl;
	BOOST_CHECK_EQUAL(it1.index1(), 1);
	BOOST_CHECK_EQUAL(it1.index2(), 0);
	++it1;
	
	std::cout << it1.index1() << ", " << it1.index2() << std::endl;
	BOOST_CHECK_EQUAL(it1.index1(), 2);
	BOOST_CHECK_EQUAL(it1.index2(), 0);
	
	
	auto it2a = it1.begin();
	
	std::cout << it2a.index1() << ", " << it2a.index2() << std::endl;
	BOOST_CHECK_EQUAL(it2a.index1(), 2);
	BOOST_CHECK_EQUAL(it2a.index2(), 0);
	++it2a;
	
	std::cout << it2a.index1() << ", " << it2a.index2() << std::endl;
	BOOST_CHECK_EQUAL(it2a.index1(), 2);
	BOOST_CHECK_EQUAL(it2a.index2(), 1);
	++it2a;
	
	std::cout << it2a.index1() << ", " << it2a.index2() << std::endl;
	BOOST_CHECK_EQUAL(it2a.index1(), 2);
	BOOST_CHECK_EQUAL(it2a.index2(), 2);
	
	
	auto it1a = it2a.begin();
	
	std::cout << it1a.index1() << ", " << it1a.index2() << std::endl;
	BOOST_CHECK_EQUAL(it1a.index1(), 2);
	BOOST_CHECK_EQUAL(it1a.index2(), 2);
	++it1a;
	
	std::cout << it1a.index1() << ", " << it1a.index2() << std::endl;
	BOOST_CHECK_EQUAL(it1a.index1(), 3);
	BOOST_CHECK_EQUAL(it1a.index2(), 2);
	++it1a;
	
	std::cout << it1a.index1() << ", " << it1a.index2() << std::endl;
	BOOST_CHECK_EQUAL(it1a.index1(), 4);
	BOOST_CHECK_EQUAL(it1a.index2(), 2);
	
}

BOOST_AUTO_TEST_CASE( prod_test )
{
	std::cout << "---------- prod test ----------" << std::endl;
	const int size1 = 10, hbw = 3;
	sym_banded_matrix<int> m1(size1, hbw, 1);
	vector<int> v1(size1, 2), v3;
	
	v3 = prod(m1, v1);
	
	std::cout << v3 << std::endl;
}

BOOST_AUTO_TEST_CASE( comparison_test )
{
	int size1 = 10, hbw = 3;
	sym_banded_matrix<double> m1(size1, hbw, 0);
	
	for(int i1 = 0; i1 < size1; ++i1){
		for(int i2 = 0; i2 <= i1; ++i2){
			if((i1 - i2) < hbw){
				m1(i1, i2) = (double)(i1 * 3 + i2);
			}
		}
	}
	
	sym_banded_matrix<double> m2(m1), m3;
	
	BOOST_CHECK_EQUAL(m1, m2);
	m2(0, 0) += 1.0;
	BOOST_CHECK(m1 != m2);
	
	m3 = m2;
	BOOST_CHECK_EQUAL(m2, m3);
	
}

